class EnvironmentSettings:
    Paths = {
        'DataBaseDir': "../../_data",
        'RawDataDir': "raw",
        'ProcessedDataDir': "processed",
        'DailyDataDir': "daily",
        'HistoricalDataDir': "historical",
        'MonthlySuffix': "monthly",
        'AnnualSuffix': "annual",
        'BhavDataDir': "bhavcopy",
        'NseDerivativesDataDir': "nse_derivatives",
        'NseIndicesDataDir': "nse_indices",
        'ManualDataPath': "../manual_data"
    }
    Development = {
        "InstrumentationLevel": 1 | 2 | 4
    }
